# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2015-2018 CERN.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Pytest configuration."""

import os
from datetime import datetime

import pytest
from flask import Blueprint, Flask
from flask_celeryext import FlaskCeleryExt
from flask_mail import Mail
from six import StringIO

from invenio_mail import InvenioMail


@pytest.fixture(scope="session")
def email_task_app(request):
    """Flask application fixture."""
    app = Flask("testapp")
    app.config.update(
        SQLALCHEMY_DATABASE_URI=os.environ.get("SQLALCHEMY_DATABASE_URI", "sqlite://"),
        CELERY_ALWAYS_EAGER=True,
        CELERY_RESULT_BACKEND="cache",
        CELERY_CACHE_BACKEND="memory",
        CELERY_EAGER_PROPAGATES_EXCEPTIONS=True,
        MAIL_SUPPRESS_SEND=True,
        MAIL_MAX_ATTACHMENT_SIZE=30,
        MAIL_MAX_RETRIES=2,
    )
    FlaskCeleryExt(app)

    InvenioMail(app, StringIO())

    return app


@pytest.fixture(scope="session")
def email_task_failure_app(request):
    """Flask application fixture."""
    app = Flask("testapp")
    app.config.update(
        SQLALCHEMY_DATABASE_URI=os.environ.get("SQLALCHEMY_DATABASE_URI", "sqlite://"),
        CELERY_ALWAYS_EAGER=True,
        CELERY_RESULT_BACKEND="cache",
        CELERY_CACHE_BACKEND="memory",
        CELERY_EAGER_PROPAGATES_EXCEPTIONS=True,
        MAIL_SUPPRESS_SEND=False,
        MAIL_MAX_ATTACHMENT_SIZE=30,
        MAIL_MAX_RETRIES=2,
    )
    FlaskCeleryExt(app)

    Mail(app)

    return app


@pytest.fixture(scope="session")
def email_api_app(email_task_app):
    """Flask application fixture."""
    email_task_app.register_blueprint(
        Blueprint("invenio_mail", __name__, template_folder="templates")
    )

    return email_task_app


@pytest.fixture
def email_params():
    """Email parameters fixture."""
    return {
        "subject": "subject",
        "recipients": ["recipient@inveniosoftware.com"],
        "sender": "sender@inveniosoftware.com",
        "cc": "cc@inveniosoftware.com",
        "bcc": "bcc@inveniosoftware.com",
        "reply_to": "reply_to@inveniosoftware.com",
        "date": datetime.now(),
        "attachments": [],
        "charset": None,
        "extra_headers": None,
        "mail_options": [],
        "rcpt_options": [],
    }


@pytest.fixture
def email_ctx():
    """Email context fixture."""
    return {
        "user": "User",
        "content": "This a content.",
        "sender": "sender",
    }


@pytest.fixture(scope="session")
def email_message():
    """Email message fixture."""
    return {
        "subject": "Subject",
        "body": "This is a test message! Do not respond!",
        "recipients": ["recipient@unit.test"],
        "sender": "sender@unit.test",
        "reply_to": "reply_to@unit.test",
    }



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2016-2025 CERN.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Configuration for Invenio-Mail.

Invenio-Mail is a tiny integration layer between Invenio and Flask-Mail, so
please refer to
`Flask-Mail <https://pythonhosted.org/Flask-Mail/#configuring-flask-mail>`_'s
list of configuration variables.
"""

import logging

MAIL_DEFAULT_REPLY_TO = None
"""Reply to mail address for e-mails."""

MAIL_MAX_ATTACHMENT_SIZE = 1000000
"""Max size of inline attachments, in bytes."""

MAIL_MAX_RETRIES = 2
"""How often will we repeat if a problem occurred."""

MAIL_MIN_LOGGING_LEVEL = logging.ERROR
"""Minimum logging level for the mail logger."""



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2024      University of Münster.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Module retry and failure tests."""

import smtplib
from unittest import mock

from celery.exceptions import Retry
from pytest import raises

from invenio_mail.tasks import _send_email_with_attachments, send_email


def test_send_email(email_task_failure_app, email_message):
    """Test if send_email is retried on failure."""
    with email_task_failure_app.app_context():
        mailer = email_task_failure_app.extensions["mail"]
        mailer.send = mock.Mock()
        mailer.send.side_effect = smtplib.SMTPHeloError(100, "Couldn't say helo!")
        with raises(Retry):
            send_email(email_message)


def test_send_email_with_attachments(email_task_failure_app, email_message):
    """Test if send_email is retried on failure."""
    with email_task_failure_app.app_context():
        mailer = email_task_failure_app.extensions["mail"]
        mailer.send = mock.Mock()
        mailer.send.side_effect = smtplib.SMTPHeloError(100, "Couldn't say helo!")
        with raises(Retry):
            _send_email_with_attachments(email_message)



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2015-2018 CERN.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Test package API."""

from invenio_mail.api import TemplatedMessage


def test_templated_message(email_api_app, email_params, email_ctx):
    """Test that all the fields given are inside the message."""
    with email_api_app.app_context():
        msg = TemplatedMessage(
            template_body="invenio_mail/base.txt",
            template_html="invenio_mail/base.html",
            ctx=email_ctx,
            **email_params
        )

        for key in email_params:
            assert email_params[key] == getattr(msg, key), key

        # let's check that the body and html are correctly formatted
        assert "<p>Dear {0},</p>".format(email_ctx["user"]) in msg.html
        assert "Dear {0},".format(email_ctx["user"]) in msg.body
        assert "<p>{0}</p>".format(email_ctx["content"]) in msg.html
        assert "{0}".format(email_ctx["content"]) in msg.body
        assert email_ctx["sender"] in msg.html
        assert email_ctx["sender"] in msg.body


def test_simple_templated_message(email_api_app):
    """Test that defaults are sane."""
    with email_api_app.app_context():
        msg = TemplatedMessage(
            template_body="invenio_mail/base.txt",
            template_html="invenio_mail/base.html",
        )
        assert msg



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2015-2025 CERN.
# Copyright (C) 2023-2024 University of Münster.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Background tasks for mail module."""

import json
import logging
import random
import smtplib
from base64 import b64decode

from celery import shared_task
from flask import current_app
from flask_mail import Message

from .errors import AttachmentOversizeException


def send_email_with_attachments(data, attachments=None):
    """Celery task for sending mails with attachments.

    :param data: a dict with the email fields
    :param attachments: a list of dicts, in the following form:
        attachments = [
                {
                    "base64": "RGllcyBpc3QgZGFzIEhhdXMgdm9tIE5pa29sYXVzCg==",
                    "disposition": "filename.bin",
                },
            ]

    Note: attachment files are inline, added as Base64 encoded strings.
    The encoded string will be part of the payload, sent over network to Celery.
    """
    if attachments is None:
        attachments = []
    for attachment in attachments:
        if len(attachment["base64"]) > current_app.config["MAIL_MAX_ATTACHMENT_SIZE"]:
            raise AttachmentOversizeException

    return _send_email_with_attachments.delay(data, attachments)


@shared_task(bind=True)
def send_email(self, data):
    """Celery task for sending emails.

    .. warning::

       Attachments do not work with Celery tasks since
       :class:`flask_mail.Attachment` is not serializable in ``JSON``
       nor ``msgpack``. Note that a
       `custom serializer <http://docs.celeryproject.org/en/latest/
       userguide/calling.html#serializers>`__
       can be created if attachments are really needed.

       This version adds attachments by putting them into a base64 encoded string.
       This is not an optimal solution, it might get problematic if they are too
       large to handle by the messaging queue, so check for a maximum size beforehand.
    """
    msg = Message()
    msg.__dict__.update(data)

    _send(msg, self, original={"data": data})


@shared_task(bind=True)
def _send_email_with_attachments(self, data, attachments=None):
    """Celery task for sending emails with attachments."""
    if attachments is None:
        attachments = []
    msg = Message()
    msg.__dict__.update(data)

    for attachment in attachments:
        rawdata = b64decode(attachment.get("base64"))
        content_type = "application/octet-stream"
        if "content_type" in attachment:
            content_type = attachment.get("content_type")
        disposition = None
        if "disposition" in attachment:
            disposition = attachment.get("disposition")
        filename = None
        if "filename" in attachment:
            filename = attachment.get("filename")
        msg.attach(
            filename=filename,
            content_type=content_type,
            data=rawdata,
            disposition=disposition,
        )

    _send(msg, self, original={"data": data, "attachments": attachments})


def _send(msg, task, original):
    """Send emails and add exception handling."""
    try:
        current_app.extensions["mail"].send(msg)
    except (
        smtplib.SMTPRecipientsRefused,
        smtplib.SMTPHeloError,
        smtplib.SMTPSenderRefused,
        smtplib.SMTPDataError,
        smtplib.SMTPNotSupportedError,
        Exception,
    ):
        if task.request.retries <= current_app.config["MAIL_MAX_RETRIES"]:
            raise task.retry(
                countdown=int(random.uniform(2, 4) ** task.request.retries)
            )
        else:
            logger = logging.getLogger("invenio-mail")
            # We log the original data as JSON, so that we can recover both the body and
            # attachments (in base64)
            logger.warning(json.dumps(original))



# -*- coding: utf-8 -*-
#
# This file is part of Invenio.
# Copyright (C) 2015-2018 CERN.
# Copyright (C) 2023      University of Münster.
#
# Invenio is free software; you can redistribute it and/or modify it
# under the terms of the MIT License; see LICENSE file for more details.

"""Module tests."""

import pytest

from invenio_mail.errors import AttachmentOversizeException
from invenio_mail.tasks import send_email, send_email_with_attachments


def test_send_message_outbox(email_task_app):
    """Test sending a message using Task module."""
    with email_task_app.app_context():
        with email_task_app.extensions["mail"].record_messages() as outbox:
            msg = {
                "subject": "Test1",
                "sender": "test1@test1.test1",
                "recipients": ["test1@test1.test1"],
            }

            send_email.delay(msg)

            assert len(outbox) == 1
            sent_msg = outbox[0]
            assert sent_msg.subject == "Test1"
            assert sent_msg.sender == "test1@test1.test1"
            assert sent_msg.recipients == ["test1@test1.test1"]


def test_send_message_stream(email_task_app):
    """Test sending a message using Task module."""
    with email_task_app.app_context():
        with email_task_app.extensions["mail"].record_messages() as outbox:
            msg = {
                "subject": "Test2",
                "sender": "test2@test2.test2",
                "recipients": ["test2@test2.test2"],
            }

            send_email.delay(msg)

            result_stream = email_task_app.extensions["invenio-mail"].stream
            assert result_stream.getvalue().find("From: test2@test2.test2") != -1
            assert result_stream.getvalue().find("To: test2@test2.test2") != -1
            assert result_stream.getvalue().find("Subject: Test2") != -1


def test_send_message_with_date(email_task_app):
    """Test sending a message with a date."""
    with email_task_app.app_context():
        msg = {
            "subject": "Test4",
            "sender": "test4@test4.test4",
            "recipients": ["test4@test4.test4"],
            "date": 1456242014.398119,
        }

        send_email.delay(msg)

        result_stream = email_task_app.extensions["invenio-mail"].stream
        assert result_stream.getvalue().find("Date: Tue, 23 Feb 2016") != -1


def test_send_message_stream_with_attachment(email_task_app):
    """Test sending a message with attachment using Task module."""
    with email_task_app.app_context():
        with email_task_app.extensions["mail"].record_messages() as outbox:
            msg = {
                "subject": "Test2",
                "sender": "test2@test2.test2",
                "recipients": ["test2@test2.test2"],
            }
            attachments = [
                {
                    "base64": "RWluIGVpbmZhY2hlciBTdHJpbmcK",
                    "disposition": "attachment",
                    "filename": "filename.bin",
                },
            ]
            send_email_with_attachments(msg, attachments)

        assert len(outbox) == 1
        msg = outbox[0].as_string()
        assert "Content-Type: application/octet-stream" in msg
        assert 'Content-Disposition: attachment; filename="filename.bin"' in msg
        assert "RWluIGVpbmZhY2hlciBTdHJpbmcK" in msg


def test_send_message_stream_with_attachment_no_filename(email_task_app):
    """Test sending a message with attachment using Task module."""
    with email_task_app.app_context():
        with email_task_app.extensions["mail"].record_messages() as outbox:
            msg = {
                "subject": "Test2",
                "sender": "test2@test2.test2",
                "recipients": ["test2@test2.test2"],
            }
            attachments = [
                {
                    "base64": "RWluIGVpbmZhY2hlciBTdHJpbmcK",
                    "disposition": "attachment",
                },
            ]
            send_email_with_attachments(msg, attachments)

        assert len(outbox) == 1
        msg = outbox[0].as_string()
        assert "Content-Type: application/octet-stream" in msg
        assert "filename=" not in msg
        assert "RWluIGVpbmZhY2hlciBTdHJpbmcK" in msg


def test_send_message_stream_with_oversize_attachment(email_task_app):
    """Test sending a message with oversize attachment."""
    with email_task_app.app_context():
        with email_task_app.extensions["mail"].record_messages():
            msg = {
                "subject": "Test2",
                "sender": "test2@test2.test2",
                "recipients": ["test2@test2.test2"],
            }
            attachments = [
                {
                    "base64": "RGllcyBpc3QgZGFzIEhhdXMgdm9tIE5pa29sYXVzCg==",
                    "disposition": "attachment",
                    "filename": "filename.bin",
                },
            ]
            with pytest.raises(AttachmentOversizeException):
                send_email_with_attachments(msg, attachments)
